//! _BndPresLimBufQ_ is a bounds-preserving, optionally limited, buffer queue.
//!
//! # Terminology
//! _length_ is used to refer the number of elements in the queue. _size_ is
//! used to refer to the total amount of bytes in the queue.
//!
//! # Example
//! ```
//! use bndpresbufq::BndPresLimBufQ;
//!
//! // Construct a queue with a maximum 2 element length limit and 4 bytes size
//! // limit
//! let mut q = BndPresLimBufQ::new(Some(2), Some(4));
//!
//! // Add elements to fill up to queue
//! q.try_push(vec![1, 2]).unwrap();
//! q.force_push(vec![3, 4]);
//!
//! // Fail to add new node
//! assert_eq!(q.try_push(vec![5]), Err(vec![5]));
//!
//! // Forcibly add node; expelling the oldest node
//! q.force_push([6].into());
//!
//! assert_eq!(q.pop(), Some(vec![3, 4]));
//! assert_eq!(q.pop(), Some(vec![6]));
//! assert_eq!(q.pop(), None);
//! ```

use limq::LimQ;

/// A bounds-preserving, optionally limited, buffer queue.
pub struct BndPresLimBufQ {
  q: LimQ<Vec<u8>>,
  max_size: Option<usize>,
  size: usize
}

impl BndPresLimBufQ {
  /// Create a new queue with an optional limit to the number of elements and
  /// total queue buffer size.
  ///
  /// Passing `None` to `max_len` will disable the explicit maximum number of
  /// queue elements limitation.  Passing `None` to `max_size` will disable
  /// the explicit maximum queue buffer size.
  ///
  /// # Panics
  /// Passing a zero to either `max_len` or `max_size` will casuse a panic.
  pub fn new(max_len: Option<usize>, max_size: Option<usize>) -> Self {
    assert!(!matches!(max_size, Some(0)));
    Self {
      q: LimQ::new(max_len),
      max_size,
      size: 0
    }
  }

  /// Change the length limit of the queue.
  ///
  /// # Panics
  /// A zero-limit will cause a panic.
  pub fn set_max_len(&mut self, max_len: Option<usize>) {
    self.q.set_max_len(max_len)
  }

  /// Change the size limit of the queue.
  ///
  /// # Panics
  /// A zero-limit will cause a panic.
  pub fn set_max_size(&mut self, max_size: Option<usize>) {
    assert!(!matches!(max_size, Some(0)));
    self.max_size = max_size;
  }

  /// Return maximum queue length.
  pub fn max_len(&self) -> Option<usize> {
    self.q.max_len()
  }

  /// Return maximum queue size.
  pub fn max_size(&self) -> Option<usize> {
    self.max_size
  }

  /// Return current number of queue elements.
  #[inline]
  pub fn len(&self) -> usize {
    self.q.len()
  }

  /// Return current total size of all buffers in queue.
  #[inline]
  pub fn size(&self) -> usize {
    self.size
  }

  /// Return boolean indicating whether the queue is empty or not.
  #[inline]
  pub fn is_empty(&self) -> bool {
    self.q.is_empty()
  }

  /// Drop elements that overflow the queue limit (if a limit has been set).
  #[inline]
  pub fn purge_overflow(&mut self) {
    self.q.purge_overflow();
    if let Some(max_size) = self.max_size {
      while self.size > max_size {
        let _ = self.q.pop();
      }
    }
  }

  /// Try to push a buffer onto the queue.
  ///
  /// Returns `Err(n)` if the queue is full (either with regards to length or
  /// size).
  pub fn try_push(&mut self, n: Vec<u8>) -> Result<(), Vec<u8>> {
    let nlen = n.len();
    if let Some(max_size) = self.max_size {
      if self.size + nlen <= max_size {
        self.q.try_push(n)?;
        self.size += nlen;
        Ok(())
      } else {
        // Would overflow (in size)
        Err(n)
      }
    } else {
      // No size limit
      self.q.try_push(n)?;
      self.size += nlen;
      Ok(())
    }
  }

  /// Forcibly add a buffer to the queue.
  ///
  /// If the input buffer `n` is larger than the configured maximum total queue
  /// size, this function returns `Err(n)`, returning the input node back to
  /// the caller.
  pub fn force_push(&mut self, n: Vec<u8>) -> Result<(), Vec<u8>> {
    let nlen = n.len();

    // If a maximum size has been set for this queue ..
    if let Some(max_size) = self.max_size {
      // .. then bail if the input buffer size is larger than the maximum
      // allowed total queue size.
      if nlen > max_size {
        return Err(n);
      }

      // Keep removing next node in line until there's room for new node
      while self.size + nlen > max_size {
        let _ = self.pop();
      }
    }

    // At this point, pushing should always succeed
    self.q.force_push(n);
    self.size += nlen;

    Ok(())
  }

  /// Take next buffer off queue.
  #[inline]
  pub fn pop(&mut self) -> Option<Vec<u8>> {
    let n = self.q.pop()?;
    self.size -= n.len();
    Some(n)
  }
}


#[cfg(test)]
mod tests {
  use super::BndPresLimBufQ;

  #[test]
  #[should_panic]
  fn zero_len() {
    let _q = BndPresLimBufQ::new(Some(0), None);
  }

  #[test]
  #[should_panic]
  fn zero_size() {
    let _q = BndPresLimBufQ::new(None, Some(0));
  }

  #[test]
  fn try_exceed_len() {
    let mut q = BndPresLimBufQ::new(Some(1), None);
    q.try_push([1].into()).unwrap();
    assert_eq!(q.len(), 1);

    let Err(e) = q.try_push([2].into()) else {
      panic!("Unexpectedly not failure");
    };
    assert_eq!(e, &[2]);

    assert_eq!(q.pop(), Some(vec![1]));
    assert_eq!(q.len(), 0);
    assert_eq!(q.pop(), None);
  }

  #[test]
  fn try_exceed_size() {
    let mut q = BndPresLimBufQ::new(None, Some(2));
    q.try_push([1].into()).unwrap();
    assert_eq!(q.len(), 1);
    assert_eq!(q.size(), 1);
    q.try_push([2].into()).unwrap();
    assert_eq!(q.len(), 2);
    assert_eq!(q.size(), 2);

    let Err(e) = q.try_push([3].into()) else {
      panic!("Unexpectedly not failure");
    };
    assert_eq!(e, vec![3]);

    assert_eq!(q.pop(), Some(vec![1]));
    assert_eq!(q.pop(), Some(vec![2]));
    assert_eq!(q.len(), 0);
    assert_eq!(q.pop(), None);
  }

  #[test]
  fn force_on_full_len() {
    let mut q = BndPresLimBufQ::new(Some(1), None);
    q.force_push([1].into()).unwrap();
    q.force_push([2].into()).unwrap();

    assert_eq!(q.pop(), Some(vec![2]));
  }

  #[test]
  fn force_on_full_size() {
    let mut q = BndPresLimBufQ::new(None, Some(2));
    q.force_push([1].into()).unwrap();
    assert_eq!(q.size(), 1);
    q.force_push([2].into()).unwrap();
    assert_eq!(q.size(), 2);
    q.force_push([3].into()).unwrap();
    assert_eq!(q.size(), 2);

    assert_eq!(q.pop(), Some(vec![2]));
    assert_eq!(q.size(), 1);
    assert_eq!(q.pop(), Some(vec![3]));

    assert!(q.is_empty());
    assert_eq!(q.size(), 0);
  }

  #[test]
  fn force_too_big_buf() {
    let mut q = BndPresLimBufQ::new(None, Some(2));
    assert_eq!(q.force_push([1, 2, 3].into()), Err(vec![1, 2, 3]));
  }
}

// vim: set ft=rust et sw=2 ts=2 sts=2 cinoptions=2 tw=79 :
